@@ -28,6 +28,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdarg.h>
+#include <signal.h>
 #include <unistd.h>
 #include <ctype.h>
 #include <time.h>
@@ -41,7 +42,6 @@
 #define EXIT_SYSDATA_ERROR -3
 #define EXIT_INVALID_ARGS  -4
 
-#define EXMFILE "/proc/exmap"
 #define BUF_SIZE 256
 
 #define CSV_COLUMNS 18
@@ -64,6 +64,8 @@ extern int rows;
 extern int columns;
 #endif
 
+static data_t data;
+
 static void
 myprintf (data_t *d, const char *fmt, ...)
 {
@@ -631,51 +633,118 @@ print_file_info (file_t * f, data_t * d)
   return lines;
 }
 
+#ifdef INTERACTIVE
+static gboolean
+srv_cmd_done (data_t * d, const char * buf)
+{
+  if (d->srv_sckt < 0)
+    return FALSE;
+  
+  if (!strncmp (buf, "EXMP_DONE", 9))
+    return TRUE;
+
+  if (!strncmp (buf, "EXMP_ERROR", 10))
+    {
+      g_debug ("Server error: %s\n", buf + 11);
+      return TRUE;
+    }
+
+  return FALSE;
+}
+#endif
+
 /* loads a list of accessible pids from /proc
  */
 static GList *
-load_pids ()
+load_pids (data_t * d)
 {
-  const gchar * n;
   GList       * pids = NULL;
-  GDir        * dir = g_dir_open ("/proc", 0, NULL);
 
-  if (!dir)
-    return NULL;
-
-  while ((n = g_dir_read_name (dir)) != NULL)
+#ifdef INTERACTIVE
+  if (d->srv_sckt >= 0)
     {
-      if (isdigit(*n))
-          pids = g_list_prepend (pids, GINT_TO_POINTER (atoi (n)));
+      char buf[256];
+      g_snprintf (buf, sizeof (buf), "EXMP_CMD SYSPIDS\n");
+      write (d->srv_sckt, buf, strlen (buf));
+
+      FILE * f = fdopen (dup (d->srv_sckt), "r");
+      
+      while (fgets (buf, sizeof (buf), f))
+        {
+#ifdef INTERACTIVE
+          if (srv_cmd_done (d, buf))
+            break;
+#endif
+          pid_t pid = atoi (buf);
+          
+          if (pid)
+            pids = g_list_prepend (pids, GINT_TO_POINTER (pid));
+        }
+
+      fclose (f);
     }
+  else
+#endif
+    {
+      const gchar * n;
+      GDir        * dir = g_dir_open ("/proc", 0, NULL);
 
-  g_dir_close (dir);
+      if (!dir)
+        return NULL;
 
+      while ((n = g_dir_read_name (dir)) != NULL)
+        {
+          if (isdigit(*n))
+            pids = g_list_prepend (pids, GINT_TO_POINTER (atoi (n)));
+        }
+
+      g_dir_close (dir);
+    }
+    
   return pids;
 }
 
+
 /* loads the contents of /proc/pid/maps for given pid into a hastable
  */
 static GHashTable *
-load_map (pid_t pid)
+load_map (data_t * d, pid_t pid)
 {
-  FILE       * fmaps = NULL;
-  gchar      * map   = g_strdup_printf ("/proc/%d/maps", pid);
-  gchar        buf[BUF_SIZE];
   GHashTable * map_hash = g_hash_table_new_full (NULL, NULL, NULL, vma_free);
   gboolean     error = FALSE;
-    
-  fmaps = fopen (map, "r");
+  FILE       * fmaps = NULL;
+  gchar        buf[BUF_SIZE];
 
-  if (!fmaps)
+#ifdef INTERACTIVE
+  if (d->srv_sckt >= 0)
     {
-      g_warning ("unable to open [%s]", map);
-      error = TRUE;
-      goto finish;
+      fmaps = fdopen (dup (d->srv_sckt), "r");
+      g_snprintf (buf, sizeof (buf), "EXMP_CMD PIDMAPS %d\n", pid);
+      write (d->srv_sckt, buf, strlen (buf));
+    }
+  else
+#endif
+    {
+      gchar * map   = g_strdup_printf ("/proc/%d/maps", pid);
+      fmaps = fopen (map, "r");
+
+      if (!fmaps)
+        {
+          g_warning ("unable to open [%s]", map);
+          error = TRUE;
+          goto finish;
+        }
+
+      g_free (map);
     }
     
-  while (fgets(buf, sizeof(buf),fmaps ))
+  while (fgets(buf, sizeof(buf), fmaps))
     {
+#ifdef INTERACTIVE
+      if (srv_cmd_done (d, buf))
+        break;
+#endif
+      
       vma_t * v = g_malloc0 (sizeof(vma_t));
       gchar * b = buf;
       gchar * p = strchr (b, '-');
@@ -807,8 +876,6 @@ load_map (pid_t pid)
   if (fmaps)
     fclose (fmaps);
 
-  g_free (map);
-
   if (error && map_hash)
     {
       g_hash_table_destroy (map_hash);
@@ -826,21 +893,21 @@ load_map (pid_t pid)
  *     for all accessible pids.
  */
 static memory_t *
-load_pid_data (pid_t pid,
+load_pid_data (data_t * d,
+               pid_t pid,
                GHashTable * map_hash,
-               GHashTable * pg_count,
-               args_t * args)
+               GHashTable * pg_count)
 {
-  FILE     * fexmp = NULL;
-  FILE     * fcmd = NULL;
   gchar      buf[BUF_SIZE];
+  gchar    * cmd_f;
   vma_t    * cur_vma  = NULL;
   gint       cur_page = 0;
-  gchar    * cmd_f;
   gboolean   error = FALSE;
   memory_t * mem_data;
+  FILE *     fexmp = NULL;
+  FILE *     fcmd = NULL;
   
-  g_return_val_if_fail (pid && map_hash && pg_count, NULL);
+  g_return_val_if_fail (d && pid && map_hash && pg_count, NULL);
   
   mem_data = g_malloc0(sizeof(memory_t));
   mem_data->pid = pid;
@@ -849,12 +916,32 @@ load_pid_data (pid_t pid,
   mem_data->files = g_hash_table_new_full (g_str_hash,g_str_equal,
                                            g_free, (GDestroyNotify)file_free);
 
-  cmd_f = g_strdup_printf ("/proc/%d/cmdline",pid);
-  fcmd = fopen (cmd_f, "r");
+#ifdef INTERACTIVE
+  if (d->srv_sckt >= 0)
+    {
+      /* remote server, we request the contents to be mirrored into
+       * our socket
+       */
+      fcmd = fdopen (dup (d->srv_sckt), "r");
+      g_snprintf (buf, sizeof (buf), "EXMP_CMD PIDCMDLINE %d\n", pid);
+      write (d->srv_sckt, buf, strlen (buf));
+    }
+  else
+#endif
+    {
+      cmd_f = g_strdup_printf ("/proc/%d/cmdline",pid);
+      fcmd = fopen (cmd_f, "r");
+      g_free (cmd_f);
+    }
+    
   if (fcmd)
     {
-      if (fgets(buf, sizeof (buf), fcmd))
+      while (fgets (buf, sizeof (buf), fcmd))
         {
+#ifdef INTERACTIVE
+          if (srv_cmd_done (d, buf))
+            break;
+#endif
           gchar * n = strchr (buf, '\n');
           if (n)
             *n = 0;
@@ -865,7 +952,7 @@ load_pid_data (pid_t pid,
            * cmd line
            */
           if (mem_data->cmd && *mem_data->cmd &&
-              args->pids_names && args->pids_names[0])
+              d->args.pids_names && d->args.pids_names[0])
             {
               /* Unfortunately, g_path_basename() does not work if the
                * cmd line has any parameters, so we extract the process name
@@ -880,7 +967,7 @@ load_pid_data (pid_t pid,
                   name = c;
                 }
       
-              char ** p = args->pids_names;
+              char ** p = d->args.pids_names;
 
               while (*p)
                 {
@@ -889,12 +976,12 @@ load_pid_data (pid_t pid,
                       /* add space for the new pid
                        * 2 == 1 for new pid 1 for NULL terminator
                        */
-                      args->pids_arg = g_realloc (args->pids_arg,
-                                                  args->pids_count + 2);
+                      d->args.pids_arg = g_realloc (d->args.pids_arg,
+                                                    d->args.pids_count + 2);
                       
-                      args->pids_arg[args->pids_count] = pid;
-                      args->pids_count++;
-                      args->pids_arg[args->pids_count] = 0;
+                      d->args.pids_arg[d->args.pids_count] = pid;
+                      d->args.pids_count++;
+                      d->args.pids_arg[d->args.pids_count] = 0;
                     }
                   
                   p++;
@@ -902,32 +989,48 @@ load_pid_data (pid_t pid,
             }
           
         }
+      
       fclose (fcmd);
     }
 
-  g_free (cmd_f);
-  
-  fexmp = fopen (EXMFILE, "w+");
-
-  if (!fexmp)
+#ifdef INTERACTIVE
+  if (d->srv_sckt >= 0)
     {
-      g_warning ("Could not open [%s]. Is exmap.ko loaded ?\n", EXMFILE);
-      exit (EXIT_KERNEL_ERROR);
+      /* remote server, we request the contents to be mirrored into
+       * our socket
+       */
+      fexmp = fdopen (dup (d->srv_sckt), "r");
+      g_snprintf (buf, sizeof (buf), "EXMP_CMD PIDDATA %d\n", pid);
+      write (d->srv_sckt, buf, strlen (buf));
     }
-
-  if (EOF == fprintf (fexmp, "%d", pid))
+  else
+#endif
     {
-      g_warning ("unable to write pid %d to [%s]",
-                 pid, EXMFILE);
+      /* local data */
+      fexmp = fopen (EXMFILE, "w+");
+      if (!fexmp)
+        {
+          /* This is an irrecoverable error, so we just exit */
+          g_warning ("Could not open [%s]. Is exmap.ko loaded ?\n", EXMFILE);
+          exit (EXIT_KERNEL_ERROR);
+        }
 
-      error = TRUE;
-      goto finish;
+      if (EOF == fprintf (fexmp, "%d", pid))
+        {
+          g_warning ("unable to write pid %d to [%s]",
+                     pid, EXMFILE);
+
+          error = TRUE;
+          goto finish;
+        }
     }
-
-  fseek (fexmp, 0, 0);
     
-  while (fgets (buf, sizeof(buf),fexmp ))
+  while (fgets (buf, sizeof(buf), fexmp))
     {
+#ifdef INTERACTIVE
+      if (srv_cmd_done (d, buf))
+        break;
+#endif
       if (!strncmp (buf, "VMA ", 4))
 	{
           gint a = strtoul (buf + 4, NULL, 16);
@@ -1029,7 +1132,6 @@ load_pid_data (pid_t pid,
     }
 
  finish:
-
   if (error && mem_data)
     {
       if (mem_data->pg_count)
@@ -1219,6 +1321,7 @@ usage (const char * option)
   fprintf (f, "  --format=which        : human|csv|csv+sums (default: human)\n");
   fprintf (f, "  --log=file            : Print results to file, not stdout\n");
   fprintf (f, "  --interactive         : Run in interactive mode\n");
+  fprintf (f, "  --remote              : Run in interactive mode for remote connection\n");
   fprintf (f, "  --version             : Program version\n");
   fprintf (f, "  --module-version      : Version of required kernel module\n");
   fprintf (f, "  --help\n\n");
@@ -1330,6 +1433,19 @@ parse_args (int argc, char ** argv, args_t * args)
           return EXIT_INVALID_ARGS;
 #endif
         }
+
+      if (!strcmp (argv[i], "--remote"))
+        {
+#ifdef INTERACTIVE
+          args->interactive = TRUE;
+          args->remote = TRUE;
+          continue;
+#else
+          g_print ("--remote not supported "
+                   "(compiled with --without-readline)\n");
+          return EXIT_INVALID_ARGS;
+#endif
+        }
       
       if (!strcmp (argv[i], "--help"))
         {
@@ -1356,9 +1472,9 @@ parse_args (int argc, char ** argv, args_t * args)
     }
 
   if (args->interactive &&
-      (args->all_pids || args->log))
+      (args->log))
     {
-      g_print ("Cannot use --all-pids or --log with --interactive.\n");
+      g_print ("Cannot use --log with --interactive.\n");
       return EXIT_INVALID_ARGS;
     }
 
@@ -1366,8 +1482,8 @@ parse_args (int argc, char ** argv, args_t * args)
   args->pids_names = g_malloc0((argc - i + 1) * sizeof(const char*));
   args->pids_count = 0;
   j = 0;
-  
-  for (; !args->all_pids && i < argc; ++i)
+
+  for (; i < argc; ++i)
     {
       args->pids_arg[args->pids_count] = atoi (argv[i]);
 
@@ -1391,6 +1507,21 @@ parse_args (int argc, char ** argv, args_t * args)
   return 0;
 }
 
+static void
+data_free (data_t * d)
+{
+  main_reset (d);
+
+  if (d->args.pids_arg)
+    g_free (d->args.pids_arg);
+
+  if (d->args.pids_names)
+    g_free (d->args.pids_names);
+  
+  fclose (d->of);
+  
+}
+
 static gboolean
 is_pid_present (pid_t * pids, pid_t pid)
 {
@@ -1405,6 +1536,12 @@ is_pid_present (pid_t * pids, pid_t pid)
   return FALSE;
 }
 
+typedef struct
+{
+  int    sckt;
+  args_t args;
+} connection_t;
+
 void
 main_load (data_t * d)
 {
@@ -1415,7 +1552,7 @@ main_load (data_t * d)
       g_print ("  Creating system snapshot ...");
     }
   
-  if (!(d->pids_sys = load_pids ()))
+  if (!(d->pids_sys = load_pids (d)))
     {
       g_warning ("There was a problem loading system data");
       exit (EXIT_SYSDATA_ERROR);
@@ -1429,13 +1566,13 @@ main_load (data_t * d)
     {
       memory_t * m;
       pid_t      p = GPOINTER_TO_INT (l->data);
-     
-      GHashTable * h = load_map (p);
+      
+      GHashTable * h = load_map (d, p);
 
       if (!h)
         continue;
-      
-      if ((m = load_pid_data (p, h, d->pg_count, &(d->args))))
+
+      if ((m = load_pid_data (d, p, h, d->pg_count)))
         {
           m->load_symbols = is_pid_present (d->args.pids_arg, p);
           d->mem_list = g_list_prepend (d->mem_list, m);
@@ -1458,6 +1595,14 @@ main_load (data_t * d)
 
       l = l->next;
     }
+
+#ifdef INTERACTIVE
+  if (d->args.interactive)
+    {
+      g_print (" %d processes running.\n", g_list_length (d->mem_list));
+      g_print ("  Enter a command or '?' for help.\n");
+    }
+#endif
 }
 
 void
@@ -1613,19 +1758,37 @@ main_reset (data_t * d)
     }
 }
 
+#ifdef INTERACTIVE
+static void
+signal_handler (int sig)
+{
+  if (sig == SIGPIPE)
+    {
+      g_print ("Connection to remote server lost.\n");
+      main_reset (&data);
+
+      if (data.srv_sckt >= 0)
+        {
+          close (data.srv_sckt);
+          data.srv_sckt = -1;
+        }
+    }
+}
+#endif
+
 int
 main (int argc, char ** argv)
 {
   gint         retval = 0;
-  data_t       data;
-  
+
   memset (&data, 0, sizeof (data));
-  
+
   data.of          = stdout;
   data.args.detail = (DETAIL_FILE_INFO | DETAIL_ELF_INFO);
   data.args.format = FORMAT_HUMAN;
-
-  if ((retval = parse_args (argc, argv, &(data.args))))
+  data.srv_sckt    = -1;
+  
+  if ((retval = parse_args (argc, argv, &data.args)))
     {
       goto finish;
     }
@@ -1637,23 +1800,30 @@ main (int argc, char ** argv)
       goto finish;
     }
 
+#ifdef INTERACTIVE
+  /* handle SIGPIPE */
+  struct sigaction sa;
+  sigfillset(&sa.sa_mask);
+  sa.sa_handler = signal_handler;
+  sigaction(SIGPIPE, &sa, NULL);
+#endif
+  
 #ifdef INTERACTIVE
   if (data.args.interactive)
     {
       interactive_init (&data);
     }
 #endif
-  
-  main_load (&data);
-  main_sort (&data);
 
+  if (!data.args.remote)
+    {
+      main_load (&data);
+      main_sort (&data);
+    }
+  
 #ifdef INTERACTIVE
   if (data.args.interactive)
     {
-      g_print (" %d processes running.\n"
-               "  Enter a command or '?' for help.\n",
-               g_list_length (data.mem_list));
-      
       interactive_loop (&data);
     }
   else
@@ -1661,15 +1831,10 @@ main (int argc, char ** argv)
     main_print (&data);
   
  finish:
-  main_reset (&data);
+  data_free (&data);
 
-  if (data.args.pids_arg)
-    g_free (data.args.pids_arg);
-
-  if (data.args.pids_names)
-    g_free (data.args.pids_names);
-  
-  fclose (data.of);
+  if (data.srv_sckt >= 0)
+    close (data.srv_sckt);
   
   return retval;
 }
